////////////////////////////////////////////////////////////////////// 
// merlin/MerlinModel.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday May 19, 2005
// 
 
#include "MerlinModel.h"
#include "MathStats.h"
#include "MerlinPDF.h"

#include <math.h>

String MerlinModel::modelsFile("param.tbl");

MerlinModel::MerlinModel()
   {
   modelCount = 0;
   models = NULL;
   }

MerlinModel::~MerlinModel()
   {
   if (models != NULL)
      delete [] models;
   }

void MerlinModel::LoadModels()
   {
   // Delete previously allocated models
   if (models != NULL)
      {
      delete [] models;
      models = NULL;
      }

   labels.Clear();
   modelCount = 0;

   // Load models file into memory
   StringArray input;
   input.Read(modelsFile);

   if (input.Length() > 0)
      {
      models = new DiseaseModel[input.Length()];

      printf("Analysis models will be retrieved from file [%s]...\n",
             (const char *) modelsFile);

      StringArray tokens, penetrances;
      for (int i = 0; i < input.Length(); i++)
         {
         input[i].Trim();

         // Skip comments
         if (input[i][0] == '#') continue;

         // Divide each line into tokens
         tokens.Clear();
         tokens.AddTokens(input[i]);

         // Skip blank lines
         if (tokens.Length() == 0) continue;

         // Print message for tracing...
         printf("   Input: %s\n", (const char *) input[i]);

         // Need a minimum of four tokens per line
         if (tokens.Length() < 4)
            {
            printf("  Status: Skipping line.\n"
                   "          Trait name, allele frequency, penetrances, and label for analysis required.\n");
            continue;
            }

         // The third token should be a set of penetrances separated by commas
         penetrances.Clear();
         penetrances.AddTokens(tokens[2], ',');
         if (penetrances.Length() != 3)
            {
            printf("  Status: Skipping line.\n"
                   "          The third token should be a set of three penetrances, separated by commas\n");
            continue;
            }

         models[modelCount].affection = Pedigree::LookupAffection(tokens[0]);

         if (models[modelCount].affection < 0)
            {
            printf("  Status: Affection %s not present in the pedigree\n", (const char *) tokens[0]);
            continue;
            }

         // First check that allele frequency is valid
         bool fail = false;

         char * ptr = NULL;
         strtod(tokens[1], &ptr);
         fail |= ptr[0] != 0;

         // Next check that penetrances are valid
         for (int j = 0; j < 3; j++)
            {
            ptr = NULL;
            strtod(penetrances[j], &ptr);
            fail |= ptr[0] != 0;
            }

         // If one of the values is not a valid number, skip
         if (fail)
            {
            printf("  Status: Invalid penetrance or allele frequency\n");
            continue;
            }

         models[modelCount].p = tokens[1];

         for (int j = 0; j < 3; j++)
            models[modelCount].pen[j] = penetrances[j];

         String label = tokens[3];

         for (int j = 4; j < tokens.Length(); j++)
            {
            label += " ";
            label += tokens[j];
            }

         printf("  Status: %s, OK \n", (const char *) label);
         labels.Push(label);

         modelCount++;
         }

      printf("Table processed. %d models recognized\n\n", modelCount);
      }
   }

int MerlinModel::CountModels()
   {
   return modelCount;
   }

AnalysisTask * MerlinModel::CreateTask()
   {
   return new ParametricAnalysis(models, labels);
   }

double ParametricAnalysis::scale = 1.0 / (2 * log(10.0));

ParametricAnalysis::ParametricAnalysis(DiseaseModel * m, StringArray & l)
   {
   modelCount = l.Length();
   labels = l;

   scores = new Vector [modelCount];
   rawScores = new Matrix [modelCount];
   impossible = new IntArray [modelCount];
   models = new DiseaseModel [modelCount];
   likelihoods = new Tree [modelCount];

   for (int i = 0; i < modelCount; i++)
      models[i].Copy(m[i]);

   baseline.Dimension(modelCount);

   scoreFile = NULL;
   }

ParametricAnalysis::~ParametricAnalysis()
   {
   delete [] likelihoods;
   delete [] models;
   delete [] scores;
   delete [] rawScores;
   }

void ParametricAnalysis::Setup(AnalysisInfo & info)
   {
   for (int i = 0; i < modelCount; i++)
      {
      scores[i].Dimension(info.positions);
      scores[i].Zero();

      impossible[i].Dimension(info.positions);
      impossible[i].Zero();

      rawScores[i].Dimension(info.positions, info.families);
      rawScores[i].Set(1.0);
      }
   }

void ParametricAnalysis::SetupFamily(AnalysisInfo & info)
   {
   LikelihoodTree tree;

   for (int i = 0; i < modelCount; i++)
      {
      tree.model.Copy(models[i]);
      tree.ScoreModel(*info.m);
      likelihoods[i].Copy(tree);

      // Check that likelihood is not zero ...
      baseline[i] = tree.WeightedSum();

      if (baseline[i] == 0.0)
         error("Disease model %s is too unlikely, family %s has zero likelihood\n",
                (const char *) labels[i], (const char *) (*info.famid));

      baseline[i] = log(baseline[i]);
      }
   }

void ParametricAnalysis::AnalyseUninformative(AnalysisInfo & info, Tree & tree)
   {
   if (scoreFile != NULL)
      for (int i = 0; i < pos; i++)
         fprintf(scoreFile, "%15s %15s %15s %10.4f\n",
                 (const char *) labels[i],
                 (const char *) info.m->family->famid,
                 (const char *) (*info.labels)[pos],
                 0.00);
   }

void ParametricAnalysis::Analyse(AnalysisInfo & info, Tree & tree, int position)
   {
   if (info.lk == -1.0)
      info.lk = tree.WeightedSum();

   double offset = -log(info.lk);

   for (int i = 0; i < modelCount; i++)
      {
      double lkPos = tree.MeanProduct(likelihoods[i]);
      double lkRatio;

      if (lkPos <= 0.0)
         {
         impossible[i][position] = true;
         rawScores[i][position][info.famno] = 1E-50;
         }
      else
         {
         lkRatio = log(lkPos) + offset - baseline[i];
         scores[i][position] += lkRatio;
         rawScores[i][position][info.famno] = exp(lkRatio);
         }

      if (scoreFile != NULL)
         {
         fprintf(scoreFile, "%15s %15s %15s ",
                 (const char *) labels[i],
                 (const char *) (*info.famid),
                 (const char *) (*info.labels)[position]);

         if (lkPos <= 0.0)
            fprintf(scoreFile, "%10s\n", "-INF");
         else
            fprintf(scoreFile, "%10.4f\n", 2 * lkRatio * scale);
         }
      }
   }

void ParametricAnalysis::ReportFamily()
   {
   for (int i = 0; i < modelCount; i++)
      likelihoods[i].Discard();
   }

void ParametricAnalysis::Report(AnalysisInfo & info)
   {
   for (model = 0; model < modelCount; model++)
      {
      if (info.drawPDF)
         {
         info.pdf->PrepareChart(4);
         info.pdf->chart.yAxis.minMin = -10.0;
         info.pdf->chart.yAxis.minMax = 4.0;
         info.pdf->chart.SetSeriesColor(0, 0.70, 0.50, 0.50);
         info.pdf->y[0].Set(0.0);
         info.pdf->chart.SetSeriesColor(1, 0.75, 0.75, 0.75);
         info.pdf->y[1].Set(3.0);
         info.pdf->chart.SetSeriesColor(2, 0.75, 0.75, 0.75);
         info.pdf->y[2].Set(-2.0);
         info.pdf->chart.SetSeriesColor(3, 0.0, 0.0, 0.0);
         info.pdf->chart.title = "Parametric Analysis for " + labels[model];
         }

      printf("Parametric Analysis, Model %s\n", (const char *) labels[model]);
      printf("=======================================================\n");
      printf("%15s %10s %10s %10s\n", "POSITION", "LOD", "ALPHA", "HLOD");

      for (pos = 0; pos < scores[model].Length(); pos++)
         {
         EstimateAlpha();

         if (impossible[model][pos])
            {
            printf("%15s %10s %10.3f %10.3f\n",
                   (const char *) (*info.labels)[pos], "-INFINITY", alpha, hlod);

            if (info.drawPDF)
               info.pdf->y[3][pos] = -999.0;
            }
         else
            {
            double lod = 2 * scores[model][pos] * scale;

            printf("%15s %10.3f %10.3f %10.3f\n",
                   (const char *) (*info.labels)[pos], lod, alpha, hlod);

            if (info.drawPDF)
               info.pdf->y[3][pos] = lod;
            }
         }

      if (info.drawPDF)
         info.pdf->DrawChart();

      printf("\n\n");
      }
   }

void ParametricAnalysis::OpenFiles(AnalysisInfo & info, String & prefix)
   {
   if (info.perFamily == false)
      return;

   filename = prefix + ".par";
   scoreFile = fopen(filename, "wt");

   if (scoreFile == NULL)
      error("Could not open file [%s] for storing results of parametric analysis\n",
            (const char *) scoreFile);

   fprintf(scoreFile, "%15s %15s %15s %10s\n", "MODEL", "FAMILY", "POSITION", "LOD");
   }

void ParametricAnalysis::CloseFiles()
   {
   if (scoreFile == NULL)
      return;

   printf("Parametric LOD scores for individual families stored in file [%s].\n",
          (const char *) filename);

   fclose(scoreFile);
   }

void ParametricAnalysis::EstimateAlpha()
   {
   // for (alpha = 0.00; alpha < 1.00; alpha += 0.05)
   //   printf("%7.3f %7.3f\n", alpha, -2.0 * f(alpha) * scale);

   // Setup range for minimization
   a = 0.0;
   c = 1.0;

   // As a starting point, assume that most families are unlinked
   b = 1.0;
   fb = f(b);

   // Call optimization routine
   Brent(0.0001);

   if (min < 1E-5 && fmin > 0.00 && fmin < 0.0001)
      alpha = hlod = 0.0;
   else
      {
      alpha = min;
      hlod = -2.0 * fmin * scale;
      }
   }

double ParametricAnalysis::f(double alpha)
   {
   double complement = 1.0 - alpha;
   double score = 0.0;

   for (int j = 0; j < rawScores[model].cols; j++)
      score += log(complement + alpha * rawScores[model][pos][j]);

   return -score;
   }

const char * ParametricAnalysis::TaskDescription()
   {
   return ("Scoring Parametric Models");
   }
 
