////////////////////////////////////////////////////////////////////// 
// libsrc/PedigreeAlleleFreq.cpp 
// (c) 2000-2005 Goncalo Abecasis
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday August 11, 2005
// 
 
#include "PedigreeAlleleFreq.h"
#include "QuickIndex.h"
#include "Error.h"

#include <math.h>

int CountAlleles(Pedigree & ped, int marker)
   {
   int result = 0;

   for (int i=0; i < ped.count; i++)
      {
      if (ped[i].markers[marker][1] > result)
         result = ped[i].markers[marker][1];
      if (ped[i].markers[marker][0] > result)
         result = ped[i].markers[marker][0];
      }

   return result;
   }

void LumpAlleles(Pedigree & ped, double threshold, int marker)
   {
   // find out how many alleles there are
   int alleles = ped.CountAlleles(marker);

   if (alleles < 3) return;

   IntArray counts(alleles + 1);
   counts.Zero();

   // Count number of occurrences for each allele
   for (int i = 0; i < ped.count; i++)
      {
      counts[ped[i].markers[marker][0]]++;
      counts[ped[i].markers[marker][1]]++;
      }

   // Calculate treshold for lumping alleles
   int total = 0;
   for (int i = 1; i <= alleles; i++)
      total += counts[i];
   int tresh = int(total * threshold);

   // Order all alleles according to their frequency
   // Zero, which corresponds to missing values, stays put!
   counts[0] = total + 1;
   QuickIndex index(counts);
   index.Reverse();

   // recode alleles
   // all alleles where frequency < tresh are labelled N
   // use counts array to keep track of labels
   int  N = 0;
   bool rare = false;
   for (int i = 0; i <= alleles; i++)
      if (counts[index[i]] > tresh)
         {
         counts[index[i]] = i;
         N++;
         }
      else
         {
         if (counts[index[i]] > 0)
            rare = true;
         counts[index[i]] = N;
         }

   for (int i = 0; i < ped.count; i++)
      {
      Alleles & current = ped[i].markers[marker];
      current[0] = counts[current[0]];
      current[1] = counts[current[1]];
      }

   MarkerInfo * info = PedigreeGlobals::GetMarkerInfo(marker);

   StringArray  oldLabels(info->alleleLabels);
   String label;

   info->alleleLabels.Clear();

   for (int i = 0; i < N; i++)
      if (oldLabels.Length() <= i)
         info->alleleLabels.Push(label = index[i]);
      else
         info->alleleLabels.Push(oldLabels[i]);

   // Reorder allele frequencies if necessary
   if (info->freq.Length())
      {
      Vector freq(info->freq);

      info->freq.Dimension(N);
      info->freq[0] = 0.0;

      for (int i = 1; i < N; i++)
         {
         info->freq[i]  = freq[index[i]];
         freq[index[i]] = 0;
         }

      // Check if we have allele frequencies for alleles not present in pedigree
      int missing_alleles = freq.CountIfGreater(0.0);

      // If such frequencies exist, they should be retained
      if (missing_alleles)
         {
         N += missing_alleles;

         info->freq.Dimension(N + 1);

         for (int i = 1, next = N - missing_alleles; i < freq.Length(); i++)
            if (freq[i] != 0.0)
               {
               if (oldLabels.Length() <= i)
                  info->alleleLabels.Push(label = index[i]);
               else
                  info->alleleLabels.Push(oldLabels[i]);
               info->freq[next++] = freq[i];
               }
         }

      if (fabs(1.0 - info->freq.Sum()) > 1e-10)
         rare = true;

      if (rare)
         {
         info->freq.Dimension(N + 1);
         info->freq[N] = 1.0 - info->freq.Sum();
         }
      }

   if (rare)
      info->alleleLabels.Push("OTHER");
   }

bool EstimateFrequencies(Pedigree & ped, int marker, int estimator)
   {
   int alleleCount = CountAlleles(ped, marker);

   IntArray founder(alleleCount + 1);
   IntArray all(alleleCount + 1);

   founder.Zero();
   all.Zero();

   for (int i = 0; i < ped.count; i++)
      {
      // When counting alleles, note that males only carry one X chromosome
      // and are arbitrarily scored as homozygous.
      all[ped[i].markers[marker][0]]++;
      if (!ped.chromosomeX || ped[i].sex != SEX_MALE)
         all[ped[i].markers[marker][1]]++;
      if (!ped[i].isFounder()) continue;
      founder[ped[i].markers[marker][0]]++;
      if (!ped.chromosomeX || ped[i].sex != SEX_MALE)
         founder[ped[i].markers[marker][1]]++;
      }

   MarkerInfo * info = ped.GetMarkerInfo(marker);

   if (info->freq.dim > 0)
      {
      // previous allele frequency information is available
      if (alleleCount >= info->freq.dim)
         error("Although marker %s has been specified with %d alleles,\n"
               "allele %d occurs in the pedigree\n",
               (const char *) info->name, info->freq.dim - 1, alleleCount);

      for (int i = 1; i <= alleleCount; i++)
         if (all[i] > 0 && info->freq[i] <= 0.0)
            error("Although allele %d for marker %s has frequency zero,\n"
                  "it occurs %d times in the pedigree",
                  i, (const char *) info->name, all[i]);

      return false;
      }
   else
      {
      if (alleleCount < 1)
         {
         // If no one is genotyped, default to two equifrequent allele
         // since some programs do not like monomorphic markers
         info->freq.Dimension(3);
         info->freq[0] = 0.0;
         info->freq[1] = 0.99999;
         info->freq[2] = 0.00001;
         return true;
         }

      info->freq.Dimension(alleleCount + 1);
      info->freq.Zero();

      if (estimator == FREQ_FOUNDERS && founder.Sum() > founder[0])
         {
         // Make sure the frequency of alleles occuring in the pedigree
         // is never zero
         for (int i = 1; i <= alleleCount; i++)
            if (founder[i] == 0 && all[i] > 0)
               founder[i] = 1;

         // To get frequencies, just multiply counts by 1 / total_counts
         double factor = 1.0 / (founder.Sum() - founder[0]);

         for (int i = 1; i <= alleleCount; i++)
            info->freq[i] = founder[i] * factor;
         }
      else if (estimator == FREQ_ALL || estimator == FREQ_FOUNDERS)
         {
         // To get frequencies, just multiply counts by 1 / total_counts
         double factor = 1.0 / (all.Sum() - all[0]);

         for (int i = 1; i <= alleleCount; i++)
            info->freq[i] = all[i] * factor;
         }
      else if (estimator == FREQ_EQUAL)
         // Assume all alleles have equal frequency
         {
         // Count the number of observed alleles
         all[0] = 0;
         int alleles = all.CountIfGreater(0);
         double freq = 1.0 / alleles;

         // Set equal frequencies for all occuring alleles
         for (int i = 0; i <= alleleCount; i++)
            info->freq[i] = all[i] ? freq : 0.0;
         }
      }

   return true;
   }

 
 
 
 
 
