////////////////////////////////////////////////////////////////////// 
// libsrc/Random.cpp 
// (c) 2000-2005 Goncalo Abecasis
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday August 11, 2005
// 
 
 
#include "Random.h"
#include "MathConstant.h"

#include <math.h>

// Constants used internally by random generator
#define  IA    16807
#define  IM    2147483647
#define  AM    (1.0 / IM)
#define  IQ    127773
#define  IR    2836
#define  NTAB  32
#define  NDIV  (1+(IM-1)/NTAB)
#define  RNMX  (1.0-EPS)

Random::Random(long s)
   {
   shuffler = new long [NTAB];
   Reset(s);
   }

void Random::Reset(long s)
   {
   // 'Continuous' Random Generator
   if ((seed = s) < 1)
      seed = s == 0 ? 1 : -s;             // seed == 0 would be disastrous
   for (int j=NTAB+7; j>=0; j--)          // Warm up and set shuffle table
      {
      long k = seed / IQ;
      seed = IA * (seed - k * IQ) - IR * k;
      if (seed < 0) seed += IM;
      if (j < NTAB) shuffler[j] = seed;
      }
   last=shuffler[0];
   normSaved = 0;

   // 'Binary' Random Generator
   bseed = last;
   }

double Random::Next()
   {
   // Compute seed = (IA * seed) % IM without overflows
   // by Schrage's method
   long k = seed / IQ;
   seed = IA * (seed - k * IQ) - IR * k;
   if (seed < 0) seed += IM;

   // Map to 0..NTAB-1
   int j = last/NDIV;

   // Output value is shuffler[j], which is in turn replaced by seed
   last = shuffler[j];
   shuffler[j] = seed;

   // Map to 0.0 .. 1.0 excluding endpoints
   double temp = AM * last;
   if  (temp > RNMX) return RNMX;
   return temp;
   }

double Random::Normal()
   {
   double v1, v2, fac, rsq;

   if (!normSaved)                   // Do we need new numbers?
      {
      do {
      v1 = 2.0 * Next() - 1.0;      // Pick two coordinates from
      v2 = 2.0 * Next() - 1.0;      // -1 to +1 and check if they
      rsq = v1*v1 + v2*v2;          // are in unit circle...
      } while (rsq >= 1.0 || rsq == 0.0);

      fac = sqrt(-2.0 * log(rsq)/rsq);    // Apply the Box-Muller
      normStore = v1 * fac;               // transformation and save
      normSaved = 1;                      // one deviate for next time
      return v2 * fac;
      }
   else
      {
      normSaved = 0;
      return normStore;
      }
   }

#define BIT1      (1 << 0)
#define BIT3      (1 << 2)
#define BIT31     (1 << 30)
#define RN_MASK   (BIT3 + BIT1)

int Random::Binary()
   {
   if (bseed & BIT31)
   // change all masked bits, shift and put 1 into first bit.
      {
      bseed = ((bseed ^ RN_MASK) << 1) | BIT1;
      return 1;
      }
   else
   // shift and put 0 onto first bit.
      {
      bseed <<= 1;
      return 0;
      }
   }

Random globalRandom;

 
 
 
 
 
