////////////////////////////////////////////////////////////////////// 
// pedstats/PedstatsHWE.cpp 
// (c) 2000-2005 Goncalo Abecasis, 2002-2005 Jan Wigginton
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Thursday August 11, 2005
// 
 
#include "PedstatsHWE.h"
#include "MathStats.h"
#include "MathConstant.h"
#include <math.h>

PedstatsHWE::PedstatsHWE(FilteredPedigree & p) : ped(p)
   {
   foundersOnly = false;
   markerTests = attemptedTests = 0;
   failedTests = pooledTests = failedPooledTests = 0;
   failedTests_01 = 0;
   significanceCutoff = 0.05;
   }

bool PedstatsHWE::TestMarker(int m, bool founders_only)
   {
   currMarker = m;
   foundersOnly = founders_only;
   hetProbs.Zero();
   poolAllele = -1;
   alleleCount = ped.CountAlleles(m);
   testPerformed = false;
   chisq = _NAN_;

   attemptedTests ++;

   if (alleleCount <= 1)
      {
      lowAlleleCountNames.Push(Person::markerNames[m]);
      return false;
      }

   InitializeArrays();
   DeterminePooledAlleleInfo(m);

   if (totalGenotypes <= 3)
      {
      sparseNames.Push(Person::markerNames[m]);
      return false;
      }

   if (realAlleles <= 1)
      {
      lowAlleleCountNames.Push(Person::markerNames[m]);
      return false;
      }

   chisq = DetermineGenotypeCounts();

   useExact = realAlleles == 2 && totalGenotypes < 5000;
   pValue = useExact ? RunExactTest() :
                    realAlleles == 1.0 ? 1.0 : chidist(chisq, realAlleles * (realAlleles - 1) * 0.5);

   markerTests ++;

   bool fail = pValue < significanceCutoff;
   pooledTests += (pooledAlleles.Sum() > 1);
   failedTests += (fail);
   failedTests_01 += pValue < 0.01;
   failedPooledTests += (fail && pooledAlleles.Sum() > 1);
   testPerformed = true;

   return fail;
   }

void PedstatsHWE::InitializeArrays()
   {
   observed.Dimension(alleleCount, alleleCount);
   observed.Zero();

   alleleFreqs.Dimension(alleleCount);
   alleleFreqs.Zero();

   pooledAlleles.Dimension(alleleCount);
   pooledAlleles.Zero();
   }

int PedstatsHWE::DetermineExpectedHeterozygotes()
   {
   double expected_hets = 0;
   int n_alleles = alleleFreqs.Length();

   for (int i = 0; i < n_alleles; i++)
      for (int j = i + 1; j < n_alleles; j++)
         expected_hets += (totalGenotypes * 2.0 * alleleFreqs[i] * alleleFreqs[j]);

   return (int) expected_hets;
   }

void PedstatsHWE::DeterminePooledAlleleInfo(int m)
   {
   smallestAllele = largestAllele = -1;
   alleleCount = alleleFreqs.Length();

   DetermineAlleleFreqs(false);
   if (totalGenotypes < 3) return;
   DetermineAlleleRange();
   PoolAlleles(alleleFreqs);
   DetermineAlleleFreqs(true);

   expectedHets = DetermineExpectedHeterozygotes();

   realAlleles = 0;
   for (int i = 0; i < alleleCount; i++)
      if (alleleFreqs[i] != 0.0)
          realAlleles ++;

   if (totalGenotypes > 3 && realAlleles > 1)
      DetermineAlleleNames();
   }

void PedstatsHWE::PoolAlleles(Vector & freqs)
   {
   // Find the minimum useful allele frequency
   double min_freq = sqrt(3.0 / (double) totalGenotypes);

   // List all rare alleles and find their total frequency
   double pooled_sum = 0.0;

   // Also find the rarest allele with frequency >= minFreq
   int mini = -1, commonCount = 0, maxi = -1;

   for (int i = 0; i < alleleCount; i++)
      if (freqs[i] == 0.0)
         continue;
      else if (freqs[i] < min_freq)
         {
         pooled_sum += freqs[i];
         pooledAlleles[i] = true;
         if (maxi == -1 || freqs[i] > freqs[maxi])
            maxi = i;
         }
      else
         {
         if (mini == -1 || freqs[i] < freqs[mini])
            mini = i;
         commonCount ++;
         }

   // If all rare alleles combined are still too rare...
   if (commonCount == 0)
      {
      pooled_sum -= freqs[maxi];
      pooledAlleles[maxi] = false;
      }
   else if (pooled_sum > 0.0 && pooled_sum < min_freq && commonCount > 1)
      {
      pooled_sum += freqs[mini];
      pooledAlleles[mini] = true;
      }

   // Glob all rare alleles into 1
   poolAllele = alleleCount - 1;
   if (pooled_sum > 0.0)
      while (pooledAlleles[poolAllele] == false)
         poolAllele--;
   }

int PedstatsHWE::DetermineAlleleFreqs(bool final_count)
   {
   totalGenotypes = 0;
   alleleFreqs.Zero();

   observed.Zero();

   for (int i = 0; i < ped.count; i++)
      {
      if (ped.filtered[i]) continue;

      if (ped[i].isGenotyped(currMarker) &&
         (!ped.chromosomeX || ped[i].sex != SEX_MALE) &&
         (!foundersOnly || ped[i].isFounder()))
         {
         int one = ped[i].markers[currMarker].one - 1;
         int two = ped[i].markers[currMarker].two - 1;

         if (final_count)
            {
            if (pooledAlleles[one]) one = poolAllele;
            if (pooledAlleles[two]) two = poolAllele;

            (one > two) ? observed[one][two] ++ : observed[two][one] ++;
            }

         alleleFreqs[one] ++;
         alleleFreqs[two] ++;
         totalGenotypes ++;
         }
      }

   if (totalGenotypes <= 3)
      return -1;

   // Work out allele frequencies
   alleleFreqs.Multiply(1.0 / (2.0 * totalGenotypes));
   }

void PedstatsHWE::DetermineAlleleRange()
   {
   for (int i = 0; i < alleleCount; i++)
      if (alleleFreqs[i] > 0)
         {
         if (smallestAllele == -1) smallestAllele = i+1;
         largestAllele = i + 1;
         }
   }

void PedstatsHWE::DetermineAlleleNames()
  {
  alleleNames.Dimension(realAlleles);

  pooledNames = "";
  int pool_ct = pooledAlleles.Sum();

  if (pool_ct > 1)
     for (int i = 0; i < alleleCount; i++)
        if (pooledAlleles[i])
           {
           pooledNames += i + 1;
           pooledNames += " ";
           }

  int j = 0;
  for (int i = 0; i < alleleCount; i++)
     {
     if (alleleFreqs[i] != 0.0)
        {
        if (i == poolAllele && pool_ct > 1)
           alleleNames[j] = "* P";
        else
           alleleNames[j].printf("%d", i+1);
        j++;
        }
     }
  }

double PedstatsHWE::DetermineGenotypeCounts()
   {
   realExpected.Dimension(realAlleles, realAlleles);
   realObserved.Dimension(realAlleles, realAlleles);

   realObserved.Zero();
   realExpected.Zero();

   double expected, chisq = 0.0;
   int alleleCount = alleleFreqs.Length();

   int l = 0, k = 0;
   for (int i = 0; i < alleleCount; i++)
      if (alleleFreqs[i] > 0)
         {
         l = 0;
         for (int j = 0; j <= i; j++)
            if (alleleFreqs[j] > 0)
               {
               expected = alleleFreqs[i] * alleleFreqs[j] * totalGenotypes * (i == j ? 1 : 2);
               chisq += square(expected - observed[i][j]) / expected;

               realExpected[k][l] = expected;
               realObserved[k][l] = observed[i][j];
               l ++;
               }
         k ++;
         }
   return chisq;
   }

double PedstatsHWE::RunExactTest()
   {
   IntArray allele;
   allele.Dimension(2);
   allele.Zero();

   int k = 0;
   for (int j = 0; j < alleleCount; j++)
      if (alleleFreqs[j] != 0)
         {
         if (k > 1)
             error("HWE::Expected two alleles, current allele is %d%d",
                   k + 1, k == 2 ? "rd" : "th");

         allele[k] = j;
         k++;
         }

   double n_rare_alleles = 2.0 * totalGenotypes;
   n_rare_alleles *= alleleFreqs[allele[0]] < alleleFreqs[allele[1]] ?
                                                  alleleFreqs[allele[0]] :
                                                  alleleFreqs[allele[1]] ;
   n_rare_alleles += 1e-6;

   if (hetProbs.Length() < n_rare_alleles + 1)
      hetProbs.Dimension((int) n_rare_alleles * 2);

   return ExactHWE(int(n_rare_alleles), int(realObserved[1][0]));
   }

double PedstatsHWE::ExactHWE(int rare , int heterozygotes)
   {
   if (rare > totalGenotypes)
      rare = 2 * totalGenotypes - rare;

   if (heterozygotes > rare)
      error("HW test: %d heterozygotes but only %d minor alleles");

   if ((rare & 1) ^(heterozygotes & 1))
      error("HW test: %s number of minor alleles, but %s number of heterozygotes",
           (rare & 1) ? "odd" : "even",
           (heterozygotes & 1) ? "odd" : "even");

   hetProbs.Dimension(rare + 1);
   hetProbs.Zero();

   // start at midpoint
   int mid = rare * (2 * totalGenotypes - rare) / (2 * totalGenotypes);

   // check to ensure that midpoint and rare alleles have same parity
   if ((rare & 1) ^ (mid & 1))
      mid++;

   int het = mid;
   int hom_r = (rare - mid) / 2;
   int hom_c = totalGenotypes - het - hom_r;

   hetProbs[mid] = 1.0;
   double sum = hetProbs[mid];

   for (het = mid; het > 1; het -= 2)
      {
      hetProbs[het - 2] = hetProbs[het] * het * (het - 1.0)
                            /(4.0 * (hom_r + 1.0) * (hom_c + 1.0));
      sum += hetProbs[het - 2];

      // 2 fewer heterozygotes for next iteration -> add one rare, one common homozygote
      hom_r++;
      hom_c++;
      }

   het = mid;
   hom_r = (rare - mid) / 2;
   hom_c = totalGenotypes - het - hom_r;

   for (het = mid; het <= rare - 2; het += 2)
      {
      hetProbs[het + 2] = hetProbs[het] * 4.0 * hom_r * hom_c
                            /((het + 2.0) * (het + 1.0));
      sum += hetProbs[het + 2];

      // add 2 heterozygotes for next iteration -> subtract one rare, one common homozygote
      hom_r--;
      hom_c--;
      }

   hetProbs /= sum;

   double p_rank = 0.0;
   for (int het = 0; het <= rare; het++)
      {
      if (hetProbs[het] > hetProbs[heterozygotes])
         continue;
      p_rank += hetProbs[het];
      }

   return min(1.0, p_rank);
   }

 
